@@ -9,21 +9,23 @@
 #include <QSpinBox>
 #include <QToolBar>
 #include <QVBoxLayout>
-#include <../UI/obs-frontend-api/obs-frontend-api.h>
+#include <obs-frontend-api.h>
 
 #include "obs-module.h"
 
 #define QT_UTF8(str) QString::fromUtf8(str)
 #define QT_TO_UTF8(str) str.toUtf8().constData()
 
-DownstreamKeyer::DownstreamKeyer(int channel, QString name)
+DownstreamKeyer::DownstreamKeyer(int channel, QString name,
+				 obs_websocket_vendor vendor)
 	: outputChannel(channel),
 	  transition(nullptr),
 	  showTransition(nullptr),
 	  hideTransition(nullptr),
 	  transitionDuration(300),
 	  showTransitionDuration(300),
-	  hideTransitionDuration(300)
+	  hideTransitionDuration(300),
+	  vendor(vendor)
 {
 	setObjectName(name);
 	auto layout = new QVBoxLayout(this);
@@ -191,23 +193,7 @@ void DownstreamKeyer::on_actionAddScene_triggered()
 	auto sceneName = QT_UTF8(obs_source_get_name(scene));
 	if (scenesList->findItems(sceneName, Qt::MatchFixedString).count() ==
 	    0) {
-		const auto item = new QListWidgetItem(sceneName);
-		scenesList->addItem(item);
-
-		std::string enable_hotkey = obs_module_text("EnableDSK");
-		enable_hotkey += " ";
-		enable_hotkey += QT_TO_UTF8(objectName());
-		std::string disable_hotkey = obs_module_text("DisableDSK");
-		disable_hotkey += " ";
-		disable_hotkey += QT_TO_UTF8(objectName());
-		uint64_t h = obs_hotkey_pair_register_source(
-			scene, enable_hotkey.c_str(), enable_hotkey.c_str(),
-			disable_hotkey.c_str(), disable_hotkey.c_str(),
-			enable_DSK_hotkey, disable_DSK_hotkey, this, this);
-
-		if (h != OBS_INVALID_HOTKEY_PAIR_ID) {
-			item->setData(Qt::UserRole, static_cast<uint>(h));
-		}
+		add_scene(sceneName, scene);
 	}
 
 	obs_source_release(scene);
@@ -265,16 +251,37 @@ void DownstreamKeyer::apply_source(obs_source_t *const newSource)
 	}
 	if (prevSource == newSource) {
 		//skip if nothing changed
-	} else if (!newTransition) {
-		obs_set_output_source(outputChannel, newSource);
 	} else {
-		obs_transition_set(newTransition, prevSource);
+		if (!newTransition) {
+			obs_set_output_source(outputChannel, newSource);
+		} else {
+			obs_transition_set(newTransition, prevSource);
 
-		obs_transition_start(newTransition, OBS_TRANSITION_MODE_AUTO,
-				     newTransitionDuration, newSource);
+			obs_transition_start(newTransition,
+					     OBS_TRANSITION_MODE_AUTO,
+					     newTransitionDuration, newSource);
 
-		if (prevTransition != newTransition)
-			obs_set_output_source(outputChannel, newTransition);
+			if (prevTransition != newTransition)
+				obs_set_output_source(outputChannel,
+						      newTransition);
+		}
+		if (vendor) {
+			const auto data = obs_data_create();
+			obs_data_set_string(data, "dsk_name",
+					    QT_TO_UTF8(objectName()));
+			obs_data_set_int(data, "dsk_channel", outputChannel);
+			obs_data_set_string(
+				data, "new_scene",
+				newSource ? obs_source_get_name(newSource)
+					  : "");
+			obs_data_set_string(
+				data, "old_scene",
+				prevSource ? obs_source_get_name(prevSource)
+					   : "");
+			obs_websocket_vendor_emit_event(
+				vendor, "dsk_scene_changed", data);
+			obs_data_release(data);
+		}
 	}
 
 	obs_source_release(prevSource);
@@ -399,7 +406,7 @@ void DownstreamKeyer::SetTransition(const char *transition_name,
 		return;
 
 	obs_source_t *newTransition = nullptr;
-	obs_frontend_source_list transitions = {0};
+	obs_frontend_source_list transitions = {};
 	obs_frontend_get_transitions(&transitions);
 	for (size_t i = 0; i < transitions.sources.num; i++) {
 		const char *n =
@@ -423,9 +430,8 @@ void DownstreamKeyer::SetTransition(const char *transition_name,
 		hideTransition = newTransition;
 	else
 		transition = newTransition;
-
-	if (oldTransition &&
-	    obs_get_output_source(outputChannel) == oldTransition) {
+	obs_source_t *prevSource = obs_get_output_source(outputChannel);
+	if (oldTransition && prevSource == oldTransition) {
 		if (newTransition) {
 			//swap transition
 			obs_transition_swap_begin(newTransition, oldTransition);
@@ -443,6 +449,7 @@ void DownstreamKeyer::SetTransition(const char *transition_name,
 			}
 		}
 	}
+	obs_source_release(prevSource);
 	if (oldTransition) {
 		obs_transition_clear(oldTransition);
 		obs_source_release(oldTransition);
@@ -630,6 +637,7 @@ void DownstreamKeyer::source_remove(void *data, calldata_t *calldata)
 bool DownstreamKeyer::enable_DSK_hotkey(void *data, obs_hotkey_pair_id id,
 					obs_hotkey_t *hotkey, bool pressed)
 {
+	UNUSED_PARAMETER(hotkey);
 	if (!pressed)
 		return false;
 	const auto downstreamKeyer = static_cast<DownstreamKeyer *>(data);
@@ -651,6 +659,7 @@ bool DownstreamKeyer::enable_DSK_hotkey(void *data, obs_hotkey_pair_id id,
 bool DownstreamKeyer::disable_DSK_hotkey(void *data, obs_hotkey_pair_id id,
 					 obs_hotkey_t *hotkey, bool pressed)
 {
+	UNUSED_PARAMETER(hotkey);
 	if (!pressed)
 		return false;
 	const auto downstreamKeyer = static_cast<DownstreamKeyer *>(data);
@@ -672,6 +681,8 @@ bool DownstreamKeyer::disable_DSK_hotkey(void *data, obs_hotkey_pair_id id,
 void DownstreamKeyer::null_hotkey(void *data, obs_hotkey_id id,
 				  obs_hotkey_t *hotkey, bool pressed)
 {
+	UNUSED_PARAMETER(id);
+	UNUSED_PARAMETER(hotkey);
 	if (!pressed)
 		return;
 	const auto downstreamKeyer = static_cast<DownstreamKeyer *>(data);
@@ -683,6 +694,8 @@ void DownstreamKeyer::null_hotkey(void *data, obs_hotkey_id id,
 bool DownstreamKeyer::enable_tie_hotkey(void *data, obs_hotkey_pair_id id,
 					obs_hotkey_t *hotkey, bool pressed)
 {
+	UNUSED_PARAMETER(id);
+	UNUSED_PARAMETER(hotkey);
 	if (!pressed)
 		return false;
 	const auto downstreamKeyer = static_cast<DownstreamKeyer *>(data);
@@ -694,6 +707,8 @@ bool DownstreamKeyer::enable_tie_hotkey(void *data, obs_hotkey_pair_id id,
 bool DownstreamKeyer::disable_tie_hotkey(void *data, obs_hotkey_pair_id id,
 					 obs_hotkey_t *hotkey, bool pressed)
 {
+	UNUSED_PARAMETER(id);
+	UNUSED_PARAMETER(hotkey);
 	if (!pressed)
 		return false;
 	const auto downstreamKeyer = static_cast<DownstreamKeyer *>(data);
@@ -705,6 +720,8 @@ bool DownstreamKeyer::disable_tie_hotkey(void *data, obs_hotkey_pair_id id,
 
 void DownstreamKeyer::AddExcludeScene(const char *scene_name)
 {
+	if (exclude_scenes.count(scene_name) > 0)
+		return;
 	exclude_scenes.emplace(scene_name);
 	const auto scene = obs_frontend_get_current_scene();
 	const auto sn = obs_source_get_name(scene);
@@ -727,6 +744,94 @@ bool DownstreamKeyer::IsSceneExcluded(const char *scene_name)
 	return exclude_scenes.find(scene_name) != exclude_scenes.end();
 }
 
+bool DownstreamKeyer::SwitchToScene(QString scene_name)
+{
+	if (scene_name.isEmpty()) {
+		on_actionSceneNull_triggered();
+		return true;
+	}
+	for (int i = 0; i < scenesList->count(); i++) {
+		const auto item = scenesList->item(i);
+		if (!item)
+			continue;
+		if (item->text() == scene_name) {
+			if (!item->isSelected()) {
+				item->setSelected(true);
+			}
+			return true;
+		}
+	}
+	return false;
+}
+
+void DownstreamKeyer::add_scene(QString scene_name, obs_source_t *s)
+{
+	const auto item = new QListWidgetItem(scene_name);
+	scenesList->addItem(item);
+
+	std::string enable_hotkey = obs_module_text("EnableDSK");
+	enable_hotkey += " ";
+	enable_hotkey += QT_TO_UTF8(objectName());
+	std::string disable_hotkey = obs_module_text("DisableDSK");
+	disable_hotkey += " ";
+	disable_hotkey += QT_TO_UTF8(objectName());
+	uint64_t h = obs_hotkey_pair_register_source(
+		s, enable_hotkey.c_str(), enable_hotkey.c_str(),
+		disable_hotkey.c_str(), disable_hotkey.c_str(),
+		enable_DSK_hotkey, disable_DSK_hotkey, this, this);
+
+	if (h != OBS_INVALID_HOTKEY_PAIR_ID) {
+		item->setData(Qt::UserRole, static_cast<uint>(h));
+	}
+}
+
+bool DownstreamKeyer::AddScene(QString scene_name)
+{
+	if (scene_name.isEmpty()) {
+		return false;
+	}
+	if (scenesList->findItems(scene_name, Qt::MatchFixedString).count() !=
+	    0) {
+		return true;
+	}
+	auto nameUtf8 = scene_name.toUtf8();
+	auto name = nameUtf8.constData();
+	auto s = obs_get_source_by_name(name);
+	if (obs_source_is_scene(s)) {
+
+		add_scene(scene_name, s);
+		obs_source_release(s);
+		return true;
+	}
+	obs_source_release(s);
+	return false;
+}
+
+bool DownstreamKeyer::RemoveScene(QString scene_name)
+{
+	if (scene_name.isEmpty()) {
+		return false;
+	}
+	for (int i = 0; i < scenesList->count(); i++) {
+		const auto item = scenesList->item(i);
+		if (!item)
+			continue;
+
+		if (item->text() == scene_name) {
+			scenesList->removeItemWidget(item);
+			obs_hotkey_pair_unregister(
+				item->data(Qt::UserRole).toUInt());
+			return true;
+		}
+	}
+	return false;
+}
+
+void DownstreamKeyer::SetTie(bool tie)
+{
+	this->tie->setChecked(tie);
+}
+
 LockedCheckBox::LockedCheckBox() {}
 
 LockedCheckBox::LockedCheckBox(QWidget *parent) : QCheckBox(parent) {}
